unit Grammar;

interface

uses Pol,ListSet;

type

  TStateOfGrammar = (_nemaNadbytecne_,_nemaNedostupne_,_jeRedukovana_,
                     _jeBezEpsilon_,_nemaCyklus_,_jeVlastni_,_jeJednoducha_,
                     _jeChomskeho_,_jeBezRekurze_,_jeGreibachova_);

  TSetOfStates = set of TStateOfGrammar;

type

  PGrammar = ^TGrammar;
  TGrammar = object

                StartSym    : PSym;
                Terminal,
                Neterminal,
                Pravidla    : PSet;
                EpsChar     : PSym;
                SetOfStates : TSetOfStates;

                constructor Init;
                destructor  Done;virtual;
                procedure   Reset;virtual;

                function  GetMemory:longint;virtual;
                function  GetNew:PGrammar;virtual;
                procedure GetCopy(g:PGrammar);virtual;
                function  GetNewCopy:PGrammar;

                procedure SetState(s:TStateOfGrammar);
                procedure ResetState(s:TStateOfGrammar);
                procedure TestStates;
                function  PlatiZe(s:TStateOfGrammar):boolean;

                procedure SetStartSym(s:PSym);
                procedure SetTerminal(t:PSet);
                procedure SetNeterminal(n:PSet);
                procedure SetPravidla(p:PSet);
                function  GetStartSym:PSym;
                function  GetTerminal:PSet;
                function  GetNeterminal:PSet;
                function  GetPravidla:PSet;

                function  Mnozina_Nt:PSet;
                function  Mnozina_Vd:PSet;
                function  Mnozina_Ne:PSet;
                function  Mnozina_Nx(x:PSym):PSet;
                function  Mnozina_Nr(x:PSym):PSet;
                function  Mnozina_Nc(x:PSym):PSet;
                function  Mnozina_Ncc(x:PSym):PSet;

                procedure KontrolaNadbytecne;
                procedure KontrolaNedostupne;
                procedure KontrolaRedukovana;
                procedure KontrolaBezEpsilon;
                procedure KontrolaCyklus;
                procedure KontrolaVlastni;
                procedure KontrolaJednoduche;
                procedure KontrolaBezRekurze;
                procedure KontrolaChomskeho;
                procedure KontrolaGreibachova;

                procedure ZrusNadbytecne;
                procedure ZrusNedostupne;
                procedure ZrusNadbytecnaPravidla;
                procedure ZrusEpsilonovaPravidla;
                procedure ZrusPravidlaSCykly;
                procedure ZrusJednoduchaPravidla;

                procedure VytvorVlastni;
                procedure VytvorChomskeho;
                procedure VytvorBezRekurze;
                procedure VytvorGreibachova;
                procedure ZjednodusSymboly;

              end;

implementation

uses Common,Rule;


{###########################  Metody TGrammar  ############################}


constructor TGrammar.Init;
begin
  if NotEnoughMemory then Fail;
  new(Terminal,Init);
  new(Neterminal,Init);
  new(Pravidla,Init);
  if  (Terminal=nil) or (Neterminal=nil) or (Pravidla=nil) then begin
    TGrammar.Done;
    Fail;
  end;
  StartSym:=Sym('');
  EpsChar:=Sym('e');
  SetOfStates:=[];
end;

destructor TGrammar.Done;
begin
  if Terminal<>nil then dispose(Terminal,Done);
  if Neterminal<>nil then dispose(Neterminal,Done);
  if Pravidla<>nil then dispose(Pravidla,Done);
end;

procedure TGrammar.Reset;
begin
  Terminal^.Reset;
  Neterminal^.Reset;
  Pravidla^.Reset;
  StartSym:=Sym('');
  EpsChar:=Sym('e');
  SetOfStates:=[];
end;

function  TGrammar.GetMemory:longint;
begin
  GetMemory:=SizeOf(TGrammar)+Terminal^.GetMemory+
             Neterminal^.GetMemory+Pravidla^.GetMemory;
end;

function  TGrammar.GetNew:PGrammar;
begin
  GetNew:=new(PGrammar,Init);
end;

procedure TGrammar.GetCopy(g:PGrammar);
begin
  if (g<>nil) and MemOk then begin
    Terminal^.GetCopy(g^.Terminal);
    Neterminal^.GetCopy(g^.Neterminal);
    Pravidla^.GetCopy(g^.Pravidla);
    g^.StartSym:=StartSym;
    g^.EpsChar:=EpsChar;
    g^.SetOfStates:=SetOfStates;
  end;
end;

function  TGrammar.GetNewCopy:PGrammar;
var p:PGrammar;
begin
  p:=GetNew;
  GetCopy(p);
  if not MemOk and (p<>nil) then begin
    dispose(p,Done);
    GetNewCopy:=nil;
  end else GetNewCopy:=p;
end;

procedure TGrammar.SetState(s:TStateOfGrammar);
begin
  SetOfStates:=SetOfStates+[s];
end;

procedure TGrammar.ResetState(s:TStateOfGrammar);
begin
  SetOfStates:=SetOfStates-[s];
end;

procedure TGrammar.TestStates;
begin
  SetOfStates:=[];
  KontrolaNadbytecne;
  KontrolaNedostupne;
  if [_nemaNadbytecne_,_nemaNedostupne_]<=SetOfStates
  then SetState(_jeRedukovana_);
  KontrolaBezEpsilon;
  KontrolaCyklus;
  if [_jeBezEpsilon_,_jeRedukovana_,_nemaCyklus_]<=SetOfStates
  then SetState(_jeVlastni_);
  KontrolaJednoduche;
  KontrolaBezRekurze;
  if _jeBezEpsilon_ in SetOfStates then begin
    KontrolaChomskeho;
    KontrolaGreibachova;
  end;
end;

function  TGrammar.PlatiZe(s:TStateOfGrammar):boolean;
begin
  PlatiZe:=s in SetOfStates;
end;

procedure TGrammar.SetStartSym(s:PSym);
begin
  StartSym:=s;
end;

procedure TGrammar.SetTerminal(t:PSet);
begin
  if t<>nil then t^.GetCopy(Terminal);
end;

procedure TGrammar.SetNeterminal(n:PSet);
begin
  if n<>nil then n^.GetCopy(Neterminal);
end;

procedure TGrammar.SetPravidla(p:PSet);
begin
  if p<>nil then p^.GetCopy(Pravidla);
end;

function  TGrammar.GetStartSym:PSym;
begin
  GetStartSym:=StartSym;
end;

function  TGrammar.GetTerminal:PSet;
begin
  GetTerminal:=Terminal;
end;

function  TGrammar.GetNeterminal:PSet;
begin
  GetNeterminal:=Neterminal;
end;

function  TGrammar.GetPravidla:PSet;
begin
  GetPravidla:=Pravidla;
end;

function  TGrammar.Mnozina_Nt:PSet;
var N:PSet;
    pom:PListSet;
    pom1:PRule;
    i:word;
    p:PItem;
    zmena:boolean;
begin
  Mnozina_Nt:=nil;
  N:=PSet(Terminal^.GetNewCopy);
  if N<>nil then begin
    N^.AddSym(Sym(epsilon));
    repeat
      zmena:=false;
      i:=1;
      p:=Pravidla^.List;
      while (i<=Pravidla^.Num) and MemOk do begin
        pom1:=PRule(p^.Item^.Pol);
        if not N^.IsInside(pom1^.Head) then begin
          pom:=PListSet(pom1^.GetNewCopy);
          if pom<>nil then begin
            N^.Prunik(pom);
            if pom^.Num=pom1^.Num then begin
              N^.AddSym(pom1^.Head);
              zmena:=true;
            end;
            dispose(pom,Done);
          end;
        end;
        inc(i);
        p:=p^.next;
      end;
    until not zmena or not MemOk;
    if MemOk then Mnozina_Nt:=N
    else dispose(N,Done);
  end;
end;

function  TGrammar.Mnozina_Vd:PSet;
var N:PSet;
    pom:PRule;
    p:PItem;
    i,m:word;
begin
  Mnozina_Vd:=nil;
  new(N,Init);
  if N<>nil then begin
    N^.AddSym(Sym(start));
    repeat
      m:=N^.Num;
      i:=1;
      p:=Pravidla^.List;
      while (i<=Pravidla^.Num) and MemOk do begin
        pom:=PRule(p^.Item^.Pol);
        if N^.IsInside(pom^.Head) then pom^.Sjednoceni(N);
        inc(i);
        p:=p^.next;
      end;
    until (m=N^.Num) or not MemOk;
    if MemOk then Mnozina_Vd:=N
    else dispose(N,Done);
  end;
end;

function  TGrammar.Mnozina_Ne:PSet;
var N:PSet;
    pom:PListSet;
    pom1:PRule;
    p:PItem;
    i:word;
    zmena:boolean;
begin
  Mnozina_Ne:=nil;
  new(N,Init);
  if N<>nil then begin
    N^.AddSym(Sym(epsilon));
    repeat
      zmena:=false;
      i:=1;
      p:=Pravidla^.List;
      while (i<=Pravidla^.Num) and MemOk do begin
        pom1:=PRule(p^.Item^.Pol);
        if not N^.IsInside(pom1^.Head) then begin
          pom:=PListSet(pom1^.GetNewCopy);
          if pom<>nil then begin
            N^.Prunik(pom);
            if pom^.Num=pom1^.Num then begin
              N^.AddSym(pom1^.Head);
              zmena:=true;
            end;
            dispose(pom,Done);
          end;
        end;
        inc(i);
        p:=p^.next;
      end;
    until not zmena or not MemOk;
    if MemOk then Mnozina_Ne:=N
    else dispose(N,Done);
  end;
end;

function  TGrammar.Mnozina_Nx(x:PSym):PSet;
var N:PSet;
    pom:PRule;
    p:PItem;
    i,m:word;
begin
  Mnozina_Nx:=nil;
  new(N,Init);
  if N<>nil then begin
    N^.AddSym(x);
    repeat
      m:=N^.Num;
      i:=1;
      p:=Pravidla^.List;
      while (i<=Pravidla^.Num) and MemOk do begin
        pom:=PRule(p^.Item^.Pol);
        with pom^ do
          if (Num=1) and N^.IsInside(Head)
          and Neterminal^.IsInside(list^.item) then
            N^.AddSym(list^.item);
        inc(i);
        p:=p^.next;
      end;
    until (m=N^.Num) or not MemOk;
    if MemOk then Mnozina_Nx:=N
    else dispose(N,Done);
  end;
end;

function  TGrammar.Mnozina_Nr(x:PSym):PSet;
var N:PSet;
    pom:PRule;
    p:PItem;
    i,m:word;
begin
  Mnozina_Nr:=nil;
  new(N,Init);
  if N<>nil then begin
    i:=1;
    p:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      pom:=PRule(p^.Item^.Pol);
      with pom^ do
        if (x=Head) and Neterminal^.IsInside(List^.Item)
        then N^.AddSym(List^.Item);
      inc(i);
      p:=p^.next;
    end;

    m:=0;
    while (m<>N^.Num) and MemOk do begin
      m:=N^.Num;
      i:=1;
      p:=Pravidla^.List;
      while (i<=Pravidla^.Num) and MemOk do begin
        pom:=PRule(p^.Item^.Pol);
        with pom^ do
          if N^.IsInside(Head) and Neterminal^.IsInside(List^.Item)
          then N^.AddSym(List^.Item);
        inc(i);
        p:=p^.next;
      end;
    end;
    if MemOk then Mnozina_Nr:=N
    else dispose(N,Done);
  end;
end;

function  TGrammar.Mnozina_Nc(x:PSym):PSet; {je v gramatice cyklus?}
var i,m:word;                               { mn. A|A=>+x}
    Nc,Ne,NNc:PSet;
    NNe,Nr:PListSet;
    r:PRule;
    p:PItem;
begin
  Mnozina_Nc:=nil;
  new(Nc,Init);
  if Nc=nil then exit;
  Ne:=Mnozina_Ne;
  if Ne<>nil then begin
    i:=1;
    p:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      r:=PRule(p^.Item^.Pol);
      if not Nc^.IsInside(r^.Head) and r^.IsInside(x) then begin
        NNe:=PListSet(r^.GetNewCopy);
        if NNe<>nil then begin
          Ne^.Rozdil(NNe);
          if (NNe^.Num=0) or
             ((NNe^.Num=1) and (PSym(NNe^.List^.Item)=x))
          then Nc^.AddSym(r^.Head);
          dispose(NNe,Done);
        end;
      end;
      inc(i);
      p:=p^.next;
    end;

    if (Nc^.Num>0) and MemOk then
      repeat
        m:=Nc^.Num;
        i:=1;
        p:=Pravidla^.List;
        while (i<=Pravidla^.Num) and MemOk do begin
          r:=PRule(p^.Item^.Pol);
          if not Nc^.IsInside(r^.Head) then begin
            NNc:=PSet(Nc^.GetNewCopy);
            if NNc<>nil then begin
              r^.Prunik(NNc);
              if NNc^.Num>0 then begin
                Nr:=PListSet(r^.GetNewCopy);
                if Nr<>nil then begin
                  Ne^.Rozdil(Nr);
                  if (Nr^.Num=0) or
                     ((Nr^.Num=1) and Nc^.IsInside(Nr^.List^.Item))
                  then Nc^.AddSym(r^.Head);
                  dispose(Nr,Done);
                end;
              end;
              dispose(NNc,Done);
            end;
          end;
          inc(i);
          p:=p^.next;
        end;
      until (m=Nc^.Num) or not MemOk;

    dispose(Ne,Done);
    if MemOk then Mnozina_Nc:=Nc
    else dispose(Nc,Done);
  end;
end;

function  TGrammar.Mnozina_Ncc(x:PSym):PSet; {trans gram. bez eps. cyklus}
var i,m:word;                               { mn. A|A=>+x}
    Nc:PSet;
    r:PRule;
    p:PItem;
begin
  Mnozina_Ncc:=nil;
  new(Nc,Init);
  if Nc<>nil then begin
    i:=1;
    p:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      r:=PRule(p^.Item^.Pol);
      if (r^.Num=1) and (PSym(r^.List^.Item)=x)
      then Nc^.AddSym(r^.Head);
      inc(i);
      p:=p^.next;
    end;

    if (Nc^.Num>0) and MemOk then
      repeat
        m:=Nc^.Num;
        i:=1;
        p:=Pravidla^.List;
        while (i<=Pravidla^.Num) and MemOk do begin
          r:=PRule(p^.Item^.Pol);
          if (r^.Num=1) and Nc^.IsInside(r^.List^.Item)
          then Nc^.AddSym(r^.Head);
          inc(i);
          p:=p^.next;
        end;
      until (m=Nc^.Num) or not MemOk;

    if MemOk then Mnozina_Ncc:=Nc
    else dispose(Nc,Done);
  end;
end;

procedure TGrammar.KontrolaNadbytecne;
var Nt,p:PSet;
begin
  Nt:=Mnozina_Nt;
  if Nt<>nil then begin
    p:=PSet(Neterminal^.GetNewCopy);
    if p<>nil then begin
      Nt^.Rozdil(p);
      if p^.Num=0 then SetState(_nemaNadbytecne_)
                  else ResetState(_nemaNadbytecne_);
      dispose(p,Done);
    end;
    dispose(Nt,Done);
  end;
end;

procedure TGrammar.KontrolaNedostupne;
var Vd,t,n:PSet;
begin
  Vd:=Mnozina_Vd;
  if Vd<>nil then begin
    t:=PSet(Terminal^.GetNewCopy);
    if t<>nil then begin
      n:=PSet(Neterminal^.GetNewCopy);
      if n<>nil then begin
        Vd^.Rozdil(n);
        Vd^.Rozdil(t);
        if (n^.Num=0) and (t^.Num=0)
        then SetState(_nemaNedostupne_)
        else ResetState(_nemaNedostupne_);
        dispose(n,Done);
      end;
      dispose(t,Done);
    end;
    dispose(Vd,Done);
  end;
end;

procedure TGrammar.KontrolaRedukovana;
begin
  KontrolaNadbytecne;
  if  PlatiZe(_nemaNadbytecne_) then begin
    KontrolaNedostupne;
    if PlatiZe(_nemaNedostupne_) then begin
      SetState(_jeRedukovana_);
      exit;
    end;
  end;
  ResetState(_jeRedukovana_);
end;

procedure TGrammar.KontrolaBezEpsilon;
var Ne:PSet;
    o:PSym;
    p:PItem;
    i:word;
    b:boolean;
begin
  Ne:=Mnozina_Ne;
  if Ne<>nil then begin
    Ne^.DelSym(Sym(epsilon));
    o:=Sym(start);
    b:=true;
    if (Ne^.Num=1) and (Ne^.IsInside(o)) then begin
      i:=1;b:=false;
      p:=Pravidla^.List;
      while not b and (i<=Pravidla^.Num) do begin
        b:=PRule(p^.Item^.Pol)^.IsInside(o);
        inc(i);
        p:=p^.next;
      end;
    end;
    if (Ne^.Num=0) or ((Ne^.Num=1) and not b)
    then SetState(_jeBezEpsilon_)
    else ResetState(_jeBezEpsilon_);
    dispose(Ne,Done);
  end;
end;

procedure TGrammar.KontrolaCyklus;
var Nc:PSet;
    x:PSym;
    p:PItem;
    i:word;
    bez:boolean;
begin
  i:=1;bez:=true;
  p:=Neterminal^.List;
  while bez and (i<=Neterminal^.Num) and MemOk do begin
    x:=PSym(p^.Item);
    Nc:=Mnozina_Nc(x);
    if Nc<>nil then begin
      if Nc^.IsInside(x) then bez:=false;
      dispose(Nc,Done);
    end;
    inc(i);
    p:=p^.next;
  end;
  if MemOk then
    if bez then SetState(_nemaCyklus_)
    else ResetState(_nemaCyklus_);
end;

procedure TGrammar.KontrolaVlastni;
begin
  KontrolaRedukovana;
  if PlatiZe(_jeRedukovana_) then begin
    KontrolaBezEpsilon;
    if PlatiZe(_jeBezEpsilon_) then begin
      KontrolaCyklus;
      if PlatiZe(_nemaCyklus_) then begin
        SetState(_jeVlastni_);
        exit;
      end;
    end;
  end;
  ResetState(_jeVlastni_);
end;

procedure TGrammar.KontrolaJednoduche;
var pom:PRule;
    p:PItem;
    i:word;
    b:boolean;
begin
  i:=1;b:=true;
  p:=Pravidla^.List;
  while b and (i<=Pravidla^.Num) do begin  {dokud nem jednoduch pravidla}
    pom:=PRule(p^.Item^.Pol);
    with pom^ do
      if (Num=1) and Neterminal^.IsInside(List^.Item) then b:=false;
    inc(i);
    p:=p^.next;
  end;
  if b then SetState(_jeJednoducha_)
  else ResetState(_jeJednoducha_);
end;

procedure TGrammar.KontrolaBezRekurze;
var Nr:PSet;
    x:PSym;
    p:PItem;
    i:word;
    bez:boolean;
begin
  i:=1;bez:=true;
  p:=Neterminal^.List;
  while bez and (i<=Neterminal^.Num) and MemOk do begin
    x:=PSym(p^.Item);
    Nr:=Mnozina_Nr(x);
    if Nr<>nil then begin
      if Nr^.IsInside(x) then bez:=false;
      dispose(Nr,Done);
    end;
    inc(i);
    p:=p^.next;
  end;
  if MemOk then
    if bez then SetState(_jeBezRekurze_)
    else ResetState(_jeBezRekurze_);
end;

procedure TGrammar.KontrolaChomskeho;
var pom:PRule;
    p:PItem;
    i:word;
    b:boolean;
begin
  if not PlatiZe(_jeBezEpsilon_) then KontrolaBezEpsilon;
  if PlatiZe(_jeBezEpsilon_) then begin
    i:=1;b:=true;
    p:=Pravidla^.List;
    while b and (i<=Pravidla^.Num) do begin
      pom:=PRule(p^.Item^.Pol);
      with pom^ do
        if (Num>2)
        or ((Num=1) and Neterminal^.IsInside(List^.Item))
        or ((Num=2)
            and not(Neterminal^.IsInside(List^.Item)
                    and Neterminal^.IsInside(List^.next^.Item)))
        then b:=false;
      inc(i);
      p:=p^.next;
    end;
    if b then begin
      SetState(_jeChomskeho_);
      exit;
    end;
  end;
  ResetState(_jeChomskeho_);
end;

procedure TGrammar.KontrolaGreibachova;
var pom:PRule;
    p,q:PItem;
    i,j:word;
    b:boolean;
begin
  if not PlatiZe(_jeBezEpsilon_) then KontrolaBezEpsilon;
  if PlatiZe(_jeBezEpsilon_) then begin
    i:=1;b:=true;
    p:=Pravidla^.List;
    while b and (i<=Pravidla^.Num) do begin
      pom:=PRule(p^.Item^.Pol);
      with pom^ do begin
        q:=List;
        if Neterminal^.IsInside(q^.Item) then b:=false;
        j:=2;
        q:=q^.next;
        while b and (j<=Num) do
          if not Neterminal^.IsInside(q^.Item) then b:=false
          else begin
            inc(j);
            q:=q^.next;
          end;
      end;
      inc(i);
      p:=p^.next;
    end;
    if b then begin
      SetState(_jeGreibachova_);
      exit;
    end;
  end;
  ResetState(_jeGreibachova_);
end;

procedure TGrammar.ZrusNadbytecne;
var Nt:PSet;
    p:PListSet;
    q:PRule;
    w:PItem;
    i:word;
begin
  Nt:=Mnozina_Nt;
  if Nt<>nil then begin
    Nt^.Prunik(Neterminal);
    i:=1;
    w:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      q:=PRule(w^.Item^.Pol);
      if Neterminal^.IsInside(q^.Head) then begin
        p:=PListSet(q^.GetNewCopy);
        if p<>nil then begin
          Nt^.Prunik(p);
          if p^.IsEqual(q) then begin
            inc(i);
            w:=w^.next;
          end else Pravidla^.DelItem(i);
          dispose(p,Done);
        end;
      end else Pravidla^.DelItem(i);
    end;
    dispose(Nt,Done);
    Neterminal^.AddSym(Sym(start));
  end;
  if MemOk then SetState(_nemaNadbytecne_);
end;

procedure TGrammar.ZrusNedostupne;
var Vd:PSet;
    p:PListSet;
    q:PRule;
    w:PItem;
    i:word;
begin
  Vd:=Mnozina_Vd;
  if Vd<>nil then begin
    Vd^.Prunik(Neterminal);
    Vd^.Prunik(Terminal);
    i:=1;
    w:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      q:=PRule(w^.Item^.Pol);
      if Neterminal^.IsInside(q^.Head) then begin
        p:=PListSet(q^.GetNewCopy);
        if p<>nil then begin
          Vd^.Prunik(p);
          if p^.IsEqual(q) then begin
            inc(i);
            w:=w^.next;
          end else Pravidla^.DelItem(i);
          dispose(p,Done);
        end;
      end else Pravidla^.DelItem(i);
    end;
    dispose(Vd,Done);
    Neterminal^.AddSym(Sym(start));
  end;
  if MemOk then SetState(_nemaNedostupne_);
end;

procedure TGrammar.ZrusNadbytecnaPravidla;
begin
  ZrusNadbytecne;
  ZrusNedostupne;
  if MemOk then SetState(_jeRedukovana_);
end;

procedure TGrammar.ZrusEpsilonovaPravidla;
var Ne,PomPrav:PSet;
    p:PRule;
    o:PSym;
    q,t:PRuleSym;
    i:word;
    PropList:TPropertyList;

  procedure Nahrad(RuleSym:PRuleSym;mnozina:PListSet;Prav:PSet);

    procedure NahradRec(n:word);
    var pom:PSym;
    begin
      if MemOk then
        with PRule(RuleSym^.Pol)^ do
          if n>Num then begin
            if (Num>1) or
               ((Num=1) and (Head<>PSym(List^.Item)) and
                (List^.Item^.GetStr<>epsilon))
            then Prav^.AddSym(RuleSym);
          end else begin
            NahradRec(n+1);
            pom:=PSym(GetItem(n));
            if mnozina^.IsInside(pom) then begin
              DelItem(n);
              NahradRec(n);
              AddItem(n,pom);
            end;
          end;
    end;

  begin
    NahradRec(1);
  end;

begin
  if PlatiZe(_jeBezEpsilon_) then exit;
  PropList.Init;
  new(PomPrav,Init);
  new(t,Init);
  Ne:=Mnozina_Ne;
  if (Ne<>nil) and (PomPrav<>nil) and (t<>nil) then
    if not Ne^.IsEmpty then begin
      while (Pravidla^.Num>0) and MemOk do begin
        q:=PRuleSym(Pravidla^.RemoveItem(1));
        p:=PRule(PRule(q^.Pol)^.GetNewCopy);
        if p<>nil then begin
          Ne^.Prunik(p);
          if not p^.IsEmpty then begin
            Nahrad(q,p,PomPrav);
            dispose(q,Done);
          end else PomPrav^.MoveSym(q);
          dispose(p,Done);
        end;
      end;

      o:=Sym(start);
      if Ne^.IsInside(o) then begin
        while (PomPrav^.Num>0) and MemOk do begin
          q:=PRuleSym(PomPrav^.RemoveItem(1));
          PRule(q^.Pol)^.Nahrad(o,StartSym);
          Pravidla^.MoveSym(q);
        end;
        o:=StartSym;
        Neterminal^.AddSym(o);
        with t^ do begin
          InitSetStr;
          SetStr(Sym(start));
          SetStr(Sym(epsilon));
          Pravidla^.AddSym(t);
          InitSetStr;
          SetStr(Sym(start));
          SetStr(o);
          Pravidla^.AddSym(t);
        end;
        SetStartSym(PropList.GetNewSym(StartSym,Neterminal,StartSym,EpsChar));
      end else
        while (PomPrav^.Num>0) and MemOk do 
          Pravidla^.MoveItem(Pravidla^.Num+1,PomPrav^.RemoveItem(1));
    end;

  if t<>nil then dispose(t,Done);
  if PomPrav<>nil then dispose(PomPrav,Done);
  if Ne<>nil then dispose(Ne,Done);
  PropList.Done;
  if MemOk then begin
    SetState(_jeBezEpsilon_);
    if PlatiZe(_jeRedukovana_) then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.ZrusPravidlaSCykly;
var Nc,PomSet,PomPrav:PSet;
    x,p:PSym;
    w,q:PItem;
    i,j,m:word;
begin
  if PlatiZe(_nemaCyklus_) then exit;
  if not PlatiZe(_jeBezEpsilon_) then ZrusEpsilonovaPravidla;
  new(PomSet,Init);
  new(PomPrav,Init);
  if (PomSet<>nil) and (PomPrav<>nil) then begin
    j:=1;
    q:=Neterminal^.List;
    while (j<=Neterminal^.Num) and MemOk do begin
      x:=PSym(q^.Item);
      Nc:=Mnozina_Ncc(x);
      if Nc<>nil then begin
        if Nc^.IsInside(x) then begin
          PomSet^.Reset;
          PomSet^.AddSym(x);
          repeat
            m:=PomSet^.Num;
            i:=1;
            w:=Pravidla^.List;
            while (i<=Pravidla^.Num) and MemOk do begin
              with PRule(w^.Item^.Pol)^ do
                if (Num=1) and PomSet^.IsInside(Head) and
                   Nc^.IsInside(List^.Item)
                then begin
                  PomSet^.AddSym(List^.Item);
                  if Head=x then DelItem(i) else inc(i);
                end else inc(i);
              w:=w^.next;
            end;
          until (m=PomSet^.Num) or not MemOk;
          PomSet^.DelSym(x);

          i:=1;
          w:=Pravidla^.List;
          while (i<=Pravidla^.Num) and MemOk do begin
            with PRule(w^.Item^.Pol)^ do
              if PomSet^.IsInside(Head) and
                 ((Num>1) or not Nc^.IsInside(List^.Item))
              then begin
                p:=Head;
                Head:=x;
                PomPrav^.AddItem(1,w^.Item);
                Head:=p;
              end;
            inc(i);
            w:=w^.next;
          end;
        end;
        dispose(Nc,Done);
      end;
      inc(j);
      q:=q^.next;
    end;

    with PomPrav^ do
      while Num>0 do
        Pravidla^.MoveSym(RemoveItem(1));
  end;

  if PomPrav<>nil then dispose(PomPrav,Done);
  if PomSet<>nil then dispose(PomSet,Done);

  if MemOk then begin
    SetState(_nemaCyklus_);
    if PlatiZe(_jeRedukovana_) then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.ZrusJednoduchaPravidla;
var Nx,PomPrav:PSet;
    pom:PRuleSym;
    x,p:PSym;
    w,q:PItem;
    i,j:word;

begin
  if PlatiZe(_jeJednoducha_) then exit;
  if not PlatiZe(_jeBezEpsilon_) then ZrusEpsilonovaPravidla;
  new(PomPrav,Init);
  if PomPrav<>nil then begin
    j:=1;
    q:=Neterminal^.List;
    while (j<=Neterminal^.Num) and MemOk do begin
      x:=PSym(q^.Item);
      inc(j);
      q:=q^.next;
      Nx:=Mnozina_Nx(x);
      if Nx<>nil then begin
        i:=1;
        w:=Pravidla^.List;
        while (i<=Pravidla^.Num) and MemOk do begin
          pom:=PRuleSym(w^.Item);
          with PRule(pom^.Pol)^ do begin
            p:=Head;
            if Nx^.IsInside(p) then
              if (Num>1) or not Neterminal^.IsInside(List^.Item)
              then begin
                Head:=x;
                PomPrav^.AddSym(pom);
                Head:=p;
              end;
            inc(i);
            w:=w^.next;
          end;
        end;
        dispose(Nx,Done);
      end;
    end;

    i:=1;
    w:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do
      with PRule(w^.Item^.Pol)^ do
        if (Num=1) and Neterminal^.IsInside(List^.Item) then
          Pravidla^.DelItem(i)
        else begin
          inc(i);
          w:=w^.next;
        end;

    with PomPrav^ do
      while Num>0 do
        Pravidla^.MoveSym(RemoveItem(1));

    dispose(PomPrav,Done);
  end;
  if MemOk then begin
    SetState(_jeJednoducha_);
    if PlatiZe(_jeRedukovana_) then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.VytvorVlastni;
begin
  if PlatiZe(_jeVlastni_) then exit;
  ResetState(_jeRedukovana_);
  if not PlatiZe(_jeBezEpsilon_) then ZrusEpsilonovaPravidla;
  if not PlatiZe(_nemaCyklus_) then ZrusPravidlaSCykly;
  ZrusNadbytecnaPravidla;
  if MemOk then SetState(_jeVlastni_);
end;

procedure TGrammar.VytvorChomskeho;
var q:PRuleSym;
    p,f,t1,t2:PItem;
    s,w:PSym;
    t,prav:PRule;
    r:string;
    i,j:word;
    red:boolean;
    setRule:PSet;
    ProperList:TPropertyList;

begin
  if PlatiZe(_jeChomskeho_) then exit;
  if PlatiZe(_jeRedukovana_) then red:=true else red:=false;
  if not PlatiZe(_jeVlastni_) then VytvorVlastni;
  if not PlatiZe(_jeJednoducha_) then ZrusJednoduchaPravidla;
  new(q,Init);
  new(setRule,Init);
  ProperList.Init;
  if (q<>nil) and (setRule<>nil) then begin
    s:=Sym('');
    t:=PRule(q^.Pol);
    t^.AddItem(1,s);
    t^.AddItem(1,s);
    t1:=t^.List;
    t2:=t^.List^.next;
    i:=1;
    p:=Pravidla^.List;
    while (i<=Pravidla^.Num) and MemOk do begin
      prav:=PRule(p^.Item^.Pol);
      if prav^.Num>1 then begin
        while (prav^.Num>=2) and MemOk do begin
          t^.Head:=prav^.Head;
          with prav^.List^ do
            if Neterminal^.IsInside(Item)
            then t1^.Item:=Item
            else t1^.Item:=ProperList.GetNewSym(PSym(Item),Neterminal,StartSym,EpsChar);
          if prav^.Num>2 then begin
            r:='';
            j:=2;
            f:=prav^.List^.next;
            while j<=prav^.Num do begin
              if f^.Item^.GetStr=start
              then r:=r+StartSym^.GetStr
              else r:=r+f^.Item^.GetStr;
              inc(j);
              f:=f^.next;
            end;
            s:=ProperList.GetNewSym(Sym(r),Neterminal,StartSym,EpsChar);
            t2^.Item:=s;
            prav^.head:=s;
          end else begin
            s:=PSym(prav^.List^.next^.Item);
            if Neterminal^.IsInside(s)
            then t2^.Item:=s
            else t2^.Item:=ProperList.GetNewSym(s,Neterminal,StartSym,EpsChar);
          end;
          setRule^.AddSym(q);
          prav^.DelItem(1);
        end;
        Pravidla^.DelItem(i);
      end else begin
        inc(i); {pravidlo pro num=1}
        p:=p^.next;
      end;
    end; {konec while pro vechna pravidla}

    setRule^.Sjednoceni(Pravidla);
    i:=1;
    t^.DelItem(1);
    s:=ProperList.GetOldNet(i);
    while (s<>nil) and MemOk do begin
      w:=ProperList.GetNewNet(i);
      if w<>nil then begin
        Neterminal^.AddSym(w);
        if Terminal^.IsInside(s) then begin
          t^.Head:=w;
          t1^.Item:=s;
          Pravidla^.AddSym(q);
        end;
      end;
      inc(i);
      s:=ProperList.GetOldNet(i);
    end;
  end;
  ProperList.Done;
  if setRule<>nil then dispose(setRule,Done);
  if q<>nil then dispose(q,Done);
  if MemOk then begin
    SetState(_jeChomskeho_);
    if red then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.VytvorBezRekurze;
var s1,PomPrav:PSet;
    q:PRule;
    qsym:PRuleSym;
    s,h:PSym;
    p:PItem;
    n,i:word;
    b,red:boolean;
    ProperList:TPropertyList;

procedure Nahrad(pr:PRuleSym);
var qpom:PRule;
    q:PRuleSym;
    s:PSym;
    n:word;
    p:PItem;
begin
  qpom:=PRule(pr^.Pol);
  s:=PSym(qpom^.RemoveItem(1));
  n:=1;p:=PomPrav^.List;
  while (n<=PomPrav^.Num) and MemOk do begin
    if s=PRule(PRuleSym(p^.Item)^.Pol)^.Head then begin
      q:=PRuleSym(pr^.GetNewCopy);
      if q<>nil then begin
        PRule(q^.Pol)^.Insert(1,PExtRule(PRuleSym(p^.Item)^.Pol));
        with PRule(q^.Pol)^ do
         if (Num>1) or (Head<>PSym(List^.Item)) then
           s1^.MoveSym(q)
         else dispose(q,Done);
      end;
    end;
    p:=p^.next;
    inc(n);
  end;
end;

begin
  if PlatiZe(_jeBezRekurze_) then exit;
  if PlatiZe(_jeRedukovana_) then red:=true else red:=false;
  if not PlatiZe(_jeVlastni_) then VytvorVlastni;
  new(s1,Init);
  if s1=nil then exit;
  new(PomPrav,Init);
  if PomPrav=nil then begin
    dispose(s1,Done);
    exit;
  end;
  ProperList.Init;

  i:=1;
  while (i<=Neterminal^.Num) and MemOk do begin
    s:=PSym(Neterminal^.GetItem(i));
    n:=1;
    while n<=Pravidla^.Num do
      if PRule(Pravidla^.GetItem(n)^.Pol)^.Head=s
      then s1^.MoveSym(Pravidla^.RemoveItem(n))
      else inc(n);
    {v s1 jsem vytvoril mnozinu pravidel zacinajich stejnym neterminalem}

    n:=1;
    p:=s1^.List;
    while (n<=s1^.Num) and MemOk do begin
      s:=PSym(PRule(p^.Item^.Pol)^.List^.Item);
      if Neterminal^.IsInside(s) and
         s^.Mensi(PRule(p^.Item^.Pol)^.Head) then begin
        Nahrad(PRuleSym(p^.Item));
        s1^.DelItem(n);
      end else begin
        p:=p^.next;
        inc(n);
      end;
    end;
    {konec uprav pravidel na primou rekurzi }

    n:=1;
    b:=false;
    p:=s1^.List;
    h:=PRule(p^.Item^.Pol)^.Head;
    while (n<=s1^.Num) and not b do
      if PSym(PRule(p^.Item^.Pol)^.List^.Item)=h then b:=true
      else begin
        inc(n);
        p:=p^.next;
      end;
    {kontrola, zda obsahuje primou rekurzi }

    while (s1^.Num>0) and MemOk do begin
      qsym:=PRuleSym(s1^.RemoveItem(1));
      if b then begin
        q:=PRule(qsym^.Pol);
        s:=PSym(q^.List^.Item);
        h:=q^.Head;
        if h^.GetStr=start then
          h:=ProperList.GetNewSym(StartSym,Neterminal,StartSym,EpsChar)
        else h:=ProperList.GetNewSym(h,Neterminal,StartSym,EpsChar);
        if s=q^.Head then begin
          q^.Head:=h;
          q^.DelItem(1);
          PomPrav^.AddSym(qsym);
          q^.AddSym(h);
          PomPrav^.MoveSym(qsym);
        end else begin
          PomPrav^.AddSym(qsym);
          q^.AddSym(h);
          PomPrav^.MoveSym(qsym);
        end;
      end else PomPrav^.MoveSym(qsym);
    end;
    inc(i);
  end;
  dispose(Pravidla,Done);
  dispose(s1,Done);
  Pravidla:=PomPrav;

  n:=1;
  s:=ProperList.GetOldNet(n);
  while s<>nil do begin
    h:=ProperList.GetNewNet(n);
    if h<>nil then
      Neterminal^.AddSym(h);
    inc(n);
    s:=ProperList.GetOldNet(n);
  end;
  ProperList.Done;
  if MemOk then begin
    SetState(_jeBezRekurze_);
    if red then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.VytvorGreibachova;
var Zrus,SetOfNr,PomPrav,Nr:PSet;
    s,h:PSym;
    q:PRuleSym;
    r,r1:PRule;
    p,p1:PItem;
    n:word;
    red:boolean;
    ProperList:TPropertyList;

 procedure Nahrad(pr:PRuleSym);
 var r:PRule;
     q:PRuleSym;
     s:PSym;
     n:word;
     p:PItem;
 begin
   r:=PRule(pr^.Pol);
   s:=PSym(r^.RemoveItem(1));
   n:=1;p:=PomPrav^.List;
   while (n<=PomPrav^.Num) and MemOk do begin
     if s=PRule(p^.Item^.Pol)^.Head then begin
       q:=PRuleSym(pr^.GetNewCopy);
       if q<>nil then begin
         PRule(q^.Pol)^.Insert(1,PRule(p^.Item^.Pol));
         PomPrav^.MoveItem(PomPrav^.Num+1,q);
         inc(n);
       end;
     end;
     inc(n);
     p:=p^.next;
   end;
 end;

 procedure Prevod(s:PSym);
 var n:word;
     p:PItem;
     q:PRuleSym;
 begin
   n:=1;p:=Pravidla^.List;
   while (n<=Pravidla^.Num) and MemOk do
     if s=PRule(p^.Item^.Pol)^.Head then begin
       q:=PRuleSym(Pravidla^.RemoveItem(n));
       if not Neterminal^.IsInside(PRule(q^.Pol)^.List^.Item) then
         PomPrav^.MoveSym(q)
       else begin
         Nahrad(q);
         dispose(q,Done);
       end;
     end else begin
       p:=p^.next;
       inc(n);
     end;
 end;

begin
  if PlatiZe(_jeGreibachova_) then exit;
  if PlatiZe(_jeRedukovana_) then red:=true else red:=false;
  if not PlatiZe(_jeVlastni_) then VytvorVlastni;
  if not PlatiZe(_jeBezRekurze_) then VytvorBezRekurze;
  ProperList.Init;
  new(SetOfNr,Init);
  new(PomPrav,Init);
  new(Zrus,Init);
  if (SetOfNr<>nil) and (PomPrav<>nil) and (Zrus<>nil) then begin
    n:=1;p:=Neterminal^.List;
    while (n<=Neterminal^.Num) and MemOk do begin   {pro kazdy neterminal}
      Nr:=Mnozina_Nr(PSym(p^.Item));           {vytvor mn neterminalu < x}
      if Nr<>nil then begin
        new(r,Init);                                       {do pravidla r}
        if r<>nil then begin
          p1:=Nr^.List;                      {dej jako extRule tu mnozinu}
          Nr^.List:=r^.List;
          with r^ do begin
            List:=p1;
            Num:=Nr^.Num;
            Head:=PSym(p^.Item);
            Pos:=1;
            Akt:=List;
          end;
          new(q,Init);
          if q<>nil then begin
            r1:=q^.Pol;
            q^.Pol:=r;
            SetOfNr^.MoveSym(q);
            dispose(r1,Done);
          end else dispose(r,Done);
        end;
        dispose(Nr,Done);                {a smaz vnitrek prazdne mn.}
      end;
      inc(n);
      p:=p^.next;
    end;
    {konec priprav na usporadani - nastavena SetOfNr}

    while (SetOfNr^.Num>0) and MemOk do begin  {pro vsechny neterminaly}
      n:=1;p:=SetOfNr^.List;
      repeat
        while (n<=SetOfNr^.Num) and
              (PRule(p^.Item^.Pol)^.Num<>0) do begin {najdi nejvetsi neterminal}
          inc(n);                          {vzdy exituje - leva rekuze}
          p:=p^.next;
        end;
        if n<=SetOfNr^.Num then begin
          s:=PRule(p^.Item^.Pol)^.Head;
          Prevod(s);                   {preved vsechna prav. zac. s do GNF}
          Zrus^.AddSym(s);
          SetOfNr^.DelItem(n);
        end;
      until (n>SetOfNr^.Num) or not MemOk;
      n:=1;p:=SetOfNr^.List;
      while (n<=SetOfNr^.Num) and MemOk do begin {aktual. nej. neterminaly}
        Zrus^.Rozdil(PRule(p^.Item^.Pol));
        inc(n);
        p:=p^.next;
      end;
      Zrus^.Reset;
    end;
    {konec usporadavani pravidel a prevodu do castecne GNF}

    while (PomPrav^.Num>0) and MemOk do begin
      q:=PRuleSym(PomPrav^.RemoveItem(1));
      r:=PRule(q^.Pol);
      n:=2;
      p:=r^.List^.next;
      while (n<=r^.Num) and MemOk do begin
        if Terminal^.IsInside(p^.Item) then
          p^.Item:=ProperList.GetNewSym(PSym(p^.Item),Neterminal,StartSym,EpsChar);
        inc(n);
        p:=p^.next;
      end;
      Pravidla^.MoveSym(q);
    end;
    {konec transformace pravidel do GNF}

    new(q,Init);
    if q<>nil then begin
      r:=PRule(q^.Pol);
      r^.AddItem(1,Sym(''));
      n:=1;
      s:=ProperList.GetOldNet(n);
      while (s<>nil) and MemOk do begin
        h:=ProperList.GetNewNet(n);
        if h<>nil then begin
          Neterminal^.AddSym(h);
          r^.Head:=h;
          r^.List^.Item:=s;
          Pravidla^.AddSym(q);
        end;
        inc(n);
        s:=ProperList.GetOldNet(n);
      end;
      dispose(q,Done);
    end;
    {konec pridani i pravidel prevodu na terminaly}

  end;
  if Zrus<>nil then dispose(Zrus,Done);
  if SetOfNr<>nil then dispose(SetOfNr,Done);
  if PomPrav<>nil then dispose(PomPrav,Done);
  ProperList.Done;
  if MemOk then begin
    SetState(_jeGreibachova_);
    if red then ZrusNadbytecnaPravidla;
  end;
end;

procedure TGrammar.ZjednodusSymboly;
var s,starts:PSym;
    m1,m2:set of char;
    p:PItem;
    i:word;
    zn,cha,ch:char;
begin
  m1:=[];m2:=[];

  with Terminal^ do begin
    p:=List;
    for i:=1 to Num do begin
      if p^.Item^.GetLength<2 then m1:=m1+[string(p^.Item^.Pol^)[1]];
      p:=p^.next;
    end;
  end;

  with Neterminal^ do begin
    p:=List;
    for i:=1 to Num do begin
      case p^.Item^.GetLength of
        1 : m1:=m1+[string(p^.Item^.Pol^)[1]];
        3 : m2:=m2+[string(p^.Item^.Pol^)[2]];
      end;
      p:=p^.next;
    end;
  end;

  case EpsChar^.GetLength of
    1 : m1:=m1+[string(EpsChar^.Pol^)[1]];
    3 : if string(EpsChar^.Pol^)[1]='<' then
          m2:=m2+[string(EpsChar^.Pol^)[2]];
  end;

  case StartSym^.GetLength of
    1 : m1:=m1+[string(StartSym^.Pol^)[1]];
    3 : m2:=m2+[string(StartSym^.Pol^)[2]];
  end;

  starts:=Sym(start);
  with Neterminal^ do
    for i:=1 to Num do begin
      s:=PSym(GetItem(i));
      if s=starts then s:=StartSym;
      if s^.GetLength>3 then begin
        ch:='$';cha:='$';
        for zn:='Z' downto 'A' do
          if not(zn in m1) then ch:=zn;
        if ch='$' then
          for zn:='z' downto 'a' do
            if not(zn in m1) then ch:=zn;
        if ch='$' then begin
          for zn:='Z' downto 'A' do
            if not(zn in m2) then cha:=zn;
          if cha='$' then
            for zn:='z' downto 'a' do
              if not(zn in m2) then cha:=zn;
        end;
        if ch<>'$' then begin
          HashTab.Vymena(s,Sym(ch));
          m1:=m1+[ch];
        end else if cha<>'$' then begin
          HashTab.Vymena(s,Sym('<'+cha+'>'));
          m2:=m2+[cha];
        end else exit;
      end;
    end;
end;

end.
